from typing import Set, List, Dict, Tuple, Optional
from enum import Enum
import enum
from copy import deepcopy


# TODO Mypy typecheck!

# TODO check auto
class Operation(Enum):
    Union = enum.auto()
    Intersection = enum.auto()


class Terminal:
    def __init__(self, name: str):  # str = "a"):
        self.name = name

    def __eq__(self, object):
        if isinstance(object, Terminal):
            return object.name == self.name
        return False

    def __hash__(self):
        return hash(self.name)

# TODO ask - only name equality?
class Nonterminal:
    def __init__(self, name: str):
        self.name = name

    def __eq__(self, object):
        if isinstance(object, Nonterminal):
            return object.name == self.name
        return False

    def __hash__(self):
        return hash(self.name)

class State:
    def __init__(self, name: str):
        self.name = name

    def __eq__(self, object):
        if isinstance(object, State):
            return object.name == self.name
        return False

    def __hash__(self):
        return hash(self.name)


class REG:
    Rules = Dict[Nonterminal, Set[Tuple[Terminal, Optional[Nonterminal]]]]

    def __init__(self, nonterminals: Set[Nonterminal],
                 terminals: Set[Terminal],
                 rules: Rules,
                 init: Nonterminal):
        self.nonterminals = nonterminals
        self.terminals = terminals
        self.rules = rules
        self.init = init


class DFA:
    Transition = Dict[Tuple[State, Terminal], State]

    # TODO ask how to write this
    def __init__(self, states: Set[State] = set(),
                 terminals: Set[Terminal] = set(),
                 transition: Transition = dict(),
                 init: State = None,
                 final: Set[State] = set()):
        self.states = states
        self.terminals = terminals
        self.transition = transition
        self.init = init
        self.final = final

    # def total(self) -> DFA:
    def total(self):
        dfa = deepcopy(self)
        # total: Transition
        total = {}

        # TODO consider import itertools
        hell_id = 0
        hell = State(str(hell_id))
        while hell in dfa.states:
            hell_id += 1
            hell.name = str(hell_id)
        dfa.states.add(hell)

        for state in dfa.states:
            for terminal in dfa.terminals:
                if (state, terminal) in dfa.transition:
                    total[state, terminal] = dfa.transition[state, terminal]
                else:
                    total[state, terminal] = hell

        dfa.transition = total
        return dfa

    def is_total(self):
        for state in self.states:
            for terminal in self.terminals:
                if (state, terminal) not in self.transition:
                    return False
        return True


    # def complement(self) -> DFA:
    def complement(self):

        # TODO is_total (also in total itself?)
        if not self.is_total():
            self.total()
        # this is not ok, total() doesn't modify self, only returns dfa

        dfa = deepcopy(self)
        new_final = set()
        for state in dfa.states:
            if state not in dfa.final:
                new_final.add(state)

        # TODO new_final = dfa.states - dfa.final

        dfa.final = new_final
        return dfa

    @staticmethod
    # def composition(dfa_1: DFA, dfa_2: DFA, operation: Operation) -> DFA:
    def composition(dfa_1, dfa_2, operation: Operation):

        # TODO ask
        if not dfa_1.is_total():
            dfa_1 = dfa_1.total()
        if not dfa_2.is_total():
            dfa_2 = dfa_2.total()

        # TODO find out - without declaration here it keeps adding to previous sets
        dfa = DFA(set(), set(), dict(), None, set())

        # new terminals
        dfa.terminals = dfa_1.terminals.union(dfa_2.terminals)
        # ATTENTION: terminals may differ -> make total at the end?

        # new states
        for state_1 in dfa_1.states:
            for state_2 in dfa_2.states:
                # f"{state_1.name}_{state_2.name}" ?
                state = State(state_1.name + "_" + state_2.name)
                dfa.states.add(state)

                # new init
                if not dfa.init and state_1 == dfa_1.init \
                        and state_2 == dfa_2.init:
                    dfa.init = state

                # new final
                if state_1 in dfa_1.final or state_2 in dfa_2.final:
                    if operation == Operation.Union:
                        dfa.final.add(state)
                    elif operation == Operation.Intersection and \
                            state_1 in dfa_1.final and \
                            state_2 in dfa_2.final:
                        dfa.final.add(state)

                # new transition
                for terminal in dfa.terminals:
                    if (state_1, terminal) in dfa_1.transition and \
                       (state_2, terminal) in dfa_2.transition:

                        move_1 = dfa_1.transition[state_1, terminal]
                        move_2 = dfa_2.transition[state_2, terminal]
                        move =  State(move_1.name + "_" + move_2.name)
                        dfa.transition[state, terminal] = move

        return dfa


# TODO ask ?
    @staticmethod
    # def union(dfa_1: DFA, dfa_2: DFA) -> DFA:
    def union(dfa_1, dfa_2):
        return DFA.composition(dfa_1, dfa_2, Operation.Union)

    @staticmethod
    # def intersection(dfa_1: DFA, dfa_2: DFA) -> DFA:
    def intersection(dfa_1, dfa_2):
        return DFA.composition(dfa_1, dfa_2, Operation.Intersection)


class NFA:
    def __init__(self, states: Set[State],
                 terminals: Set[Terminal],
                 transition: Dict[Tuple[State, Terminal], Set[State]],
                 init: State,
                 final: Set[State]):
        self.states = states
        self.terminals = terminals
        self.transition = transition
        self.init = init
        self.final = final
