from parser import Parser
from objects import Terminal, Nonterminal, State, DFA, REG
from objects import Operation as op


def test_dfa():
    q_0 = State("q_0")
    q_1 = State("q_1")
    a = Terminal("a")
    states = {q_0, q_1}
    init = q_0
    final = {q_0}
    terminals = {a}
    transition = dict()
    transition[q_0, a] = q_1
    transition[q_1, a] = q_1
    dfa = DFA(states, terminals, transition, init, final)

    return dfa

def test_dfa2():
    r_0 = State("r_0")
    r_1 = State("r_1")
    r_2 = State("r_2")
    a = Terminal("a")
    b = Terminal("b")
    states = {r_0, r_1, r_2}
    init = r_0
    final = {r_2}
    terminals = {a, b}
    transition = dict()
    transition[r_0, a] = r_1
    transition[r_0, b] = r_1
    transition[r_1, a] = r_2
    transition[r_1, b] = r_2
    dfa = DFA(states, terminals, transition, init, final)

    return dfa

def test_reg():
    S = Nonterminal("S")
    A = Nonterminal("A")
    a = Terminal("a")
    nonterminals = {S, A}
    init = S
    terminals = {a}
    rules = dict()
    rules[S] = {(a, S), (a)}
    rules[A] = {(a)}
    reg = REG(nonterminals, terminals, rules, init)

    return reg

def test_reg2():
    S = Nonterminal("S")
    A = Nonterminal("A")
    a = Terminal("a")
    nonterminals = {S, A}
    init = S
    terminals = {a}
    rules = dict()
    reg = REG(nonterminals, terminals, rules, init)

    return reg




def main():
    dfa = test_dfa()
    dfa2 = test_dfa2()
    reg = test_reg()
    reg2 = test_reg2()
    parser = Parser()

    print("test: dfa_to_str")
    print(parser.dfa_to_str(dfa))
    print()
    print(parser.dfa_to_str(dfa, True))
    print()

    print("test: add terminal, make total, is_total")
    b = Terminal("b")
    dfa.terminals.add(b)
    print(dfa.is_total())
    total_dfa = dfa.total()
    print(parser.dfa_to_str(total_dfa, True))
    print(total_dfa.is_total())
    print()

    print("test: complement")
    compl_dfa = dfa.complement()
    print(parser.dfa_to_str(compl_dfa, True))
    print()

    print("test: reg_to_str")
    print(parser.reg_to_str(reg))
    print()
    print(parser.reg_to_str(reg, True))
    print()

    print(parser.reg_to_str(reg2))
    print()
    print(parser.reg_to_str(reg2, True))
    print()

    print("test: second dfa_to_str")
    print(parser.dfa_to_str(dfa2, True))
    print()

    print("test: composition")
    union = dfa.composition(dfa, dfa2, op.Union)
    print(parser.dfa_to_str(union, True))
    print()
    intersection = dfa.composition(dfa, dfa2, op.Intersection)
    print(parser.dfa_to_str(intersection, True))
    print()

main()